##    threadskane. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com
    

import random
from typing import Any, Callable, List

Aggregator = Callable

class Generator:
    def __init__(self):
        pass
    
    def generate(self):
        pass
    
class ChoiceGenerator(Generator):
    def __init__(self, options:List[Any] = []):
        super().__init__()
        self.options = options
        
    def generate(self):
        if len(self.options) == 0:
            return None
        return random.choice(self.options)
0    
class RangeGenerator(Generator):
    def __init__(self, maxValue: int = 255, minValue: int = 0):
        super().__init__() 
        self.maxValue = maxValue
        self.minValue = minValue
    
    def generate(self):
        return random.randint(self.minValue, self.maxValue)

class ComposedGenerator(Generator):
    def __init__(self, aggregator:Callable, generators:List[Generator] = []):
        self.aggregator = aggregator
        self.generators = generators
        super().__init__()
        
    def generate(self):
        return self.aggregator([g.generate() for g in self.generators])
    

class Style:
    def __init__(self, styleName:str, styleValue:str) -> None:
        self.styleName = styleName
        self.styleValue = styleValue
        
    def __str__(self):
        return f'\t{self.styleName}: {self.styleValue};'

class StyleRule:
    def __init__(self, selector:str, styles:List[Style] = []):
        self.selector = selector
        self.styles = styles

class StyleSheet:
    def __init__(self, styleRules:List[StyleRule] = []):
        self.styleRules = styleRules
        
    def __str__(self) -> str:
        result = ''
        for styleRule in self.styleRules:
            result += styleRule.selector + '{\n'
            for style in styleRule.styles:
                result += str(style) + '\n'
            result += '}\n\n'
        return result


def compose_color(function:str = 'rbg', components:List[int] = [0,0,0]):
    if len(components) <= 2:
        return 'black'
    else:
        return f'{function}({components[0]}, {components[1]}, {components[2]})'

def create_numeric_choice_generator(median:int = 127, delta:int = 10):
    return ChoiceGenerator([i for i in range(median-delta, median+delta)])
    
shortLinearGenerator = RangeGenerator()
shortLinearGenerators = [
    create_numeric_choice_generator(50, 20),
    create_numeric_choice_generator(130, 40),
    create_numeric_choice_generator(200, 20),
]

fonts = [
    'candara',
    '"times new roman"',
    'arial',
    '"century gothic"'
]

decorations = [
    'none'
]

rgbGenerator = ComposedGenerator(lambda x:compose_color('rgb', x), [random.choice(shortLinearGenerators) for i in range(3)])
lightRGBGenerator = ComposedGenerator(lambda x:compose_color('rgb', x), [shortLinearGenerators[0] for i in range(3)])
intenseRGBGenerator = ComposedGenerator(lambda x:compose_color('rgb', x), [shortLinearGenerators[2] for i in range(3)])
fontGenerator = ChoiceGenerator(fonts)



styles = {
    'color': intenseRGBGenerator,
    'background': lightRGBGenerator,
    'font-family': fontGenerator,
    'text-decoration': ChoiceGenerator(decorations)
}

def random_style(selector:str):
    global styles
    styles = [Style(i, styles[i].generate()) for i in styles]
    return StyleRule(selector, styles)

def random_stylesheet(selectors:List[str] = []):
    styleRules = [random_style(selector) for selector in selectors]
    return StyleSheet(styleRules)


##    threadskane. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

from functools import reduce
import json
from re import L
import re
import socket
from sys import maxsize
from threading import Thread
import threading
import time
import uuid
import os
from enum import Enum, IntEnum


class RedirectType(IntEnum):
    MOVED_PERMANENT = 301
    PERMANENT_REDIRECT = 308
    FOUND = 302
    SEE_OTHER = 303
    TEMPORARY_REDIRECT = 307
    MULTIPLE_CHOICE = 301
    NOT_MODIFIED = 304

def get_absolute_path(fileName):
    return os.sep.join([os.sep.join(__file__.split(os.sep)[:-1]), fileName])

responseCodes = {}
threadsnakeIcon = ''

with open(get_absolute_path('response_codes.txt'), 'r') as f:
    for l in [i.strip() for i in f.readlines()]:
        values = l.split('\t')
        if len(values) > 1:
            responseCodes[int(values[0])] = values[1]
        else:
            print(f'Invalid line on response_codes.txt: {values}')

with open(get_absolute_path('icon.txt'), 'rb') as f:
    threadsnakeIcon = f.read().decode('ANSI')


contentTypes = {
    "text/": ["html", "css"],
    "text/javascript": ["js"],
    "text/html": ["htm"],
    "application/": ["json", "xml", "pdf", "exe"],
    "image/": ["gif", "png", "jpeg", "bmp", "webp"],
    "image/jpeg": ["jpg"],
    "image/x-icon": ["ico"],
    "audio/": ["mpeg", "webm", "ogg", "midi", "wav"],
    "text/plain": ["txt", "*"]
}

def get_content_type(path:str):
    contentType = 'text/plain'
    if '.' in path:
        extension:str = path.split(".")[-1:][0]
        for i in contentTypes:
            if extension in contentTypes[i]:
                contentType = i
                if contentType.endswith("/"):
                    contentType += extension
                break
    return contentType

def decode_querystring(data):
    data = data.replace('+', ' ').replace('%20', ' ')#space
    data = data.replace('%2B', '+')#space
    data = data.replace('%7E', '~')#space
    references = [(i, chr(int(re.findall(r'[\d]+', i)[0]))) for i in re.findall(r'&#[\d]+;', data)]
    for ref in references:
        data = data.replace(ref[0], ref[1])
    return data

def map_dictionary(data, rowSeparator, keySeparator):
    return dict([
        tuple([j.strip() for j in i.split(keySeparator, 1)])
        for i in data.split(rowSeparator) if len(i.split(keySeparator)) == 2
    ])

class HttpResponse:
    def __init__(self):
        self.resposeStatus = 404
        self.resposeText = 'NotFound'
        self.headers = {
            "Content-Type": "text/plain",
            "Content-Length": "0",
            "Connection": "close"
        }
        self.cookieHeaders = []
        self.body = ""
        self.ended = False
        self.contentDisposition = ''
        self.cacheValue = None
        self.encoding = None

    def set_encoding(self, encoding:str):
        self.encoding = encoding
        return self

    def cache(self, cacheValue):
        self.cacheValue = cacheValue

    def end(self, data, status = None):
        self.ended = True
        self.status(status or 200)
        return self.write(data)

    def redirect(self, url, statusCode = None):
        self.status(statusCode or RedirectType.TEMPORARY_REDIRECT)
        self.headers = {"Location": url}
        self.body = ""
        return self

    def status(self, resposeStatus, resposeText=None):
        self.resposeStatus = resposeStatus
        if resposeText == None and resposeStatus in responseCodes:
            self.resposeText = responseCodes[resposeStatus]
        else:
            self.resposeText = resposeText or "OK"
        return self

    def content_type(self, value):
        self.headers["Content-Type"] = value
        return self

    def json(self, data):
        self.body = json.dumps(data)
        return self.content_type("text/json")

    def write(self, data):
        self.body += data
        return self

    def html(self, data):
        return self.content_type('text/html').write(str(data))

    def read_file(self, fileName, contentType=None):
        with open(fileName, 'r') as f:
            self.body += f.read()
        self.status(200)
        return self.content_type(contentType or get_content_type(fileName))

    def transmit_as_file(self, fileName, data, contentType=None, encoding = None):
        self.set_content_disposition('attachment', fileName).set_encoding(encoding or 'UTF-8')
        self.body = data
        self.status(200)
        return self.content_type(contentType or get_content_type(fileName))

    def set_content_disposition(self, contentDisposition, fileName = None):
        self.contentDisposition = contentDisposition
        if fileName is not None:
            self.contentDisposition += f'; fileName="{fileName}"'
        return self

    def download_file(self, path, fileName, contentType=None, encoding = None):
        self.set_content_disposition('attachment', fileName).set_encoding(encoding or 'UTF-8')
        self.body = ""
        return self.read_file(path, contentType)

    def __str__(self):
        if not self.cacheValue == None:
            return self.cacheValue
        result = f"HTTP/1.1 {self.resposeStatus} {self.resposeText}\n"
        if 'Location' not in self.headers: self.headers["Content-Length"] = len(self.body)
        if len(self.contentDisposition or '') > 0:
            self.headers['Content-Disposition'] = self.contentDisposition 
        mappedHeaders = [f"{i}:{self.headers[i]}\n" for i in self.headers]
        for cookie in self.cookieHeaders:
            print(cookie)
            mappedHeaders.append(f"Set-Cookie:{cookie}\n")
        result += reduce(lambda a, b: a + b, mappedHeaders)
        result += '\n'
        result += self.body
        return result

    def set_cookie(self, name, value, durationSec = None, domain = None, path = None):
        cookieString = value
        if durationSec != None: 
            expireUTC = time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.gmtime(time.time() + durationSec))
            cookieString += f"; Expires={expireUTC}"
        if domain != None: cookieString += f"; Domain={domain}"
        if path != None: cookieString += f"; Path={path}"
        self.cookieHeaders.append(f"{name}={cookieString}")

class HttpRequest:
    def __init__(self, raw, address):
        self.clientAddress = address
        self.load(raw)

    def load(self, raw):
        self.raw = raw
        self.authorization = {}
        self.files = {}
        headerAndBody = self.raw.split('\r\n\r\n', maxsplit=1)
        self.headers = headerAndBody[0].split('\r\n')
        if len(headerAndBody) > 1:
            self.body = headerAndBody[1]
        else:
            self.body = ''
        firstLine = self.headers[:1][0].split(' ')
        self.headers = self.headers[1:]
        self.method = firstLine[0]
        self.httpVersion = firstLine[2] if len(firstLine) >= 2 else 'HTTP/1.1'
        self.url = firstLine[1] if len(firstLine) >= 1 else ''
        self.querystring = self.url.split('?', 1)[1] if '?' in self.url else ''
        self.querystring = decode_querystring(self.querystring)
        self.path = self.url.split('?')[:1][0]
        self.baseUrl = self.url if '?' not in self.url else self.url.split('?', 1)[0]
        self.params = map_dictionary(self.querystring, '&', '=')
        self.data = {}
        self.headers = dict([
            tuple([j.strip() for j in i.split(':', 1)]) for i in self.headers
            if len(i.split(':')) == 2
        ])
        #print(self.headers)
        self.contentType = self.headers.get('Content-Type', None)
        self.cookies = {}
        if 'Cookie' in self.headers:
            self.cookies = dict([tuple([j.strip() for j in i.split('=')]) for i in self.headers['Cookie'].split(';') if len(i.split('=')) == 2])
        pass



class ServerWorker(Thread):
    def __init__(self, action):
        Thread.__init__(self)
        self.action = action
    
    def run(self):
        self.action()

class Server(Thread):
    def __init__(self, port=80, connectionTimeout=None):
        Thread.__init__(self)
        self.connectionTimeout = connectionTimeout or 0.1
        self.maxPacket = 32768
        self.port = port
        self.server_active = True

    def stop(self):
        if self.serverSocket != None:
            self.server_active = False
            srv = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            srv.connect(('localhost', self.port))
            srv.shutdown(socket.SHUT_RDWR)
            srv.close()
            self.serverSocket = None

    def __exit__(self):
        self.stop()

    def __del__(self):
        self.stop()

    def onConnect(self, clientPort, clientAddress):
        pass

    def onReceive(self, clientPort, data):
        raise NotImplementedError()

    def receive(self, clientPort):
        rdata = []
        timeout = clientPort.gettimeout()
        try:
            clientPort.settimeout(self.connectionTimeout)
            while True:
                try:    
                    rdata.extend(clientPort.recv(self.maxPacket))
                except:
                    break
        finally:
            clientPort.settimeout(timeout)
        with open('datab.txt', 'wb') as f:
            f.write(bytes(rdata))
        raw = bytes(rdata).decode('latin1')#''.join([i for i in rdata]) #I'm pretty scared about it. Decoding with ANSI is the most recent change
        #print(raw)
        return raw#''.join([i + '\n' for i in raw.splitlines()])

    def next_free(self, port, max_port=65535):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        while port <= max_port:
            try:
                sock.bind(('', port))
                sock.close()
                return port
            except OSError:
                port += 1
        raise IOError('no free ports')

    def run(self):
        self.port = self.next_free(self.port)
        print(f'listening on {self.port}...')
        self.serverSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.serverSocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 0)
        self.serverSocket.bind(('', self.port))
        self.serverSocket.listen(5)
        while self.server_active:
            try:
                #print('ON Accept')
                (clientSocket, address) = self.serverSocket.accept()
                if not self.server_active:
                    break
                #print('ACCEPT ENDED')
                def handler():
                    localSocketRef = clientSocket
                    self.onConnect(localSocketRef, address)
                    try:
                        raw_data = self.receive(localSocketRef)
                        self.onReceive(localSocketRef, raw_data, address)
                    except socket.timeout:
                        print('timeout')
                    except Exception as e:
                        print('Exception in handler')
                        print(e)
                        pass
                    finally:
                        try:
                            localSocketRef.close()
                        except:
                            print('Error closing the socket...')
                ServerWorker(handler).start()
            except OSError as o:
                print('There was an OSError!')
                self.stop()
                break
            except Exception as e:
                print('Exception in listen loop')
                print(e)
                self.stop()
                break
        #print('block ended!')

class Session:
    def __init__(self, cookieName = None):
        self.cookieName = cookieName or 'sessionId'
        self.sessions = {}

    def ensure_session(self, req:HttpRequest, res:HttpResponse, reset:bool = False):
        sessionId = ''
        if self.cookieName not in req.cookies:
            sessionId = str(uuid.uuid4())
            res.set_cookie(self.cookieName, sessionId)
            req.cookies[self.cookieName] = sessionId
        else:
            sessionId = req.cookies[self.cookieName]
        if sessionId not in self.sessions or reset:
            self.sessions[sessionId] = {}
        return sessionId

    def get(self, req:HttpRequest, res:HttpResponse, key):
        sessionId = self.ensure_session(req, res)
        if key not in self.sessions[sessionId]:
            return None
        else:
            return self.sessions[sessionId][key]

    def set(self, req:HttpRequest, res:HttpResponse, key, value):
        sessionId = self.ensure_session(req, res)
        self.sessions[sessionId][key] = value

    def create_session(self, req:HttpRequest, res:HttpResponse):
        return RequestSession(req, res, self)

class RequestSession:
    def __init__(self, req:HttpRequest, res:HttpResponse, session:Session):
        self.req = req
        self.res = res
        self.session = session
    
    def __setitem__(self, key, value):
        self.session.set(self.req, self.res, key, value)

    def __getitem__(self, key):
        return self.session.get(self.req, self.res, key)
    
    def try_get(self, key, defaultValue = None):
        value = self.session.get(self.req, self.res, key)
        if value is None:
            value = defaultValue
            self.session.set(self.req, self.res, key, value)
        return value
    
    def reset_session(self):
        self.session.ensure_session(self.req, self.res, True)


import setuptools
import os

def readfile(filename):
    with open(filename, 'r', encoding='utf-8') as f:
        return f.read()

def writefile(filename, data):
    with open(filename, 'w', encoding='utf-8') as f:
        f.write(data)

def get_version():
    versionFile = os.path.join(os.path.dirname(__file__), 'version.txt')
    data = readfile(versionFile).split('.')
    currentVersion = '.'.join([data[0], data[1], str(int(data[2]) + 1)])
    writefile(versionFile, currentVersion)
    return currentVersion

setuptools.setup(    
    name="threadsnake",
    version="0.0.3",
    author="Erick Fernando Mora Ramirez",
    author_email="erickfernandomoraramirez@gmail.com",
    description="A tiny experimental server-side express-like library",
    long_description=readfile("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/codeRookieErick/thread-snake",
    project_urls={
        "Bug Tracker": "https://dev.moradev.dev/threadsnake/issues/",
        "Documentation": "https://dev.moradev.dev/threadsnake/documentation/",
        "Examples": "https://dev.moradev.dev/threadsnake/examples/",
    },
    package_data={
        "":["*.txt"]
    },
    classifiers=[
    "Programming Language :: Python :: 3",
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
)


##    threadskane. A tiny experimental server-side express-like library.
##    Copyright (C) 2022  Erick Fernando Mora Ramirez
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <https://www.gnu.org/licenses/>.
##
##    mailto:erickfernandomoraramirez@gmail.com

from cgitb import reset
from dataclasses import replace
from functools import reduce
import json
from sys import path
from types import FunctionType
from typing import Any, Callable, Dict, List, Union

from .myhttp import Server, HttpRequest, HttpResponse, Session, RequestSession
from base64 import b64decode
from uuid import uuid4
import os
import re
import time
from importlib import __import__ as importpy
import importlib.util



'''
Loads a library from specified path.
Thanks to 
[[https://stackoverflow.com/users/7779/sebastian-rittau|Sebastian Rittau|target="_blank"]] and 
[[https://stackoverflow.com/users/3907364/raven|Raven|target="_blank"]]
'''
def import_library(name, path):
    spec = importlib.util.spec_from_file_location(name, path)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    return mod

class RouterTemplate:
    def __init__(self) -> None:
        self.R:Router = None 

def load_router(name, path) -> RouterTemplate:
    return import_library(name, path)

def routes_to(name:str):
    path = name if name.endswith('.py') else name + '.py'
    libName = 'lib'+str(uuid4()).replace('-', '')
    return load_router(libName, path).R

def load_routers(folder:str = 'routers', onExcept:Callable = None):
    result = {}
    for f in [(i[:-3], os.sep.join(folder, i)) for i in os.listdir(folder) if i.endswith('.py')]:
        try:
            result = result[f[0]] = load_router(f[0], f[1]).R
        except Exception as e:
            if onExcept is not None:
                onExcept(e)
    return result

class Router:
    def __init__(self):
        #print('router created')
        self.routes = {}
        self.callbackMutator:CallbackMutator = lambda a: a

    def register_function(self, httpMethod:str, route:str):
        ref:Router = self
        def inner(function:Callback) -> Callback:
            function = self.callbackMutator(function)
            if httpMethod.upper() not in ref.routes:
                ref.routes[httpMethod.upper()] = {route: function}
            else:
                ref.routes[httpMethod.upper()][route] = function
            return function
        return inner
    
    def use_globally(self, mutator):
        callbackMutator = self.callbackMutator
        def innerMutator(c):
            return mutator(callbackMutator(c))
        self.callbackMutator = innerMutator
        return self

    def get(self, route):
        return self.register_function('GET', route)

    def post(self, route):
        return self.register_function('POST', route)
    
    def put(self, route):
        return self.register_function('PUT', route)
    
    def delete(self, route):
        return self.register_function('DELETE', route)

    def configure(self, handler):
        self.stack.append(handler)
        return self

    def use_router(self, router, root):
        for method in router.routes:
            for action in router.routes[method]:
                self.register_function(method,f'{root}{action}')(router.routes[method][action])
                #if method not in self.routes:
                #    self.routes[method] = {}
                #self.routes[method][f'{root}{action}'] = router.routes[method][action]
        return self

    def serve_statically(self, route:str, content:Union[str, Callable[[Any, HttpRequest, HttpResponse],str]], encoding:str=None):
        def sub(s:str, ss:str)->str:
            return s if ss not in s else s[s.index(ss)+1:]
        fileName = sub(sub(route, '/'), '\\')
        def callback(app:Application, req:HttpRequest, res:HttpResponse):
            data = content if isinstance(content, str) else content(app, req, res)
            res.transmit_as_file(fileName, data, encoding=encoding)
        self.get(route)(callback)
        return self

    def __getattr__(self, method):
        def inner(route):
            return self.register_function(method.upper(), route)
        return inner


class Application(Server, Router):
    def __init__(self, port, connectionTimeout=None):
        Server.__init__(self, port, connectionTimeout)
        Router.__init__(self)
        self.stack = []
        self.session:RequestSession = None
        
    def cli_loop(self, main_event: Callable):
        self.start()
        continueLoop = True
        while(continueLoop):
            try:
                continueLoop = main_event()
            except KeyboardInterrupt as e:
                print('[CLI EXIT]')
                break
            finally:
                self.stop()
                
    def wait_exit(self, message:str = 'press [Enter] to exit...'):
        try:
            self.start()
            input(message)
        except:
            pass
        finally:
            self.stop()

    def onReceive(self, clientPort, data, clientAddress):
        if(len(data) == 0): return
        req = None
        res = HttpResponse()
        try:
            req = HttpRequest(data, clientAddress)
        except Exception as e:
            print(e)
            print(data)
            res.status(403, "BadRequest")
            clientPort.send(str(res).encode())
            return

        stack = self.stack.copy()

        def next():
            if len(stack) > 0 and not res.ended:
                stack.pop()(self, req, res, next)
        
        if ':' in req.path:
            pathParts = req.path.split('/')
            newPath = []
            for i in pathParts:
                queryPassParam = i.split(':')
                if len(queryPassParam) == 2:
                    req.params[queryPassParam[0]] = queryPassParam[1]
                else:
                    newPath.append(i)
            req.path = '/'.join(newPath)

        if req.method in self.routes:
            regularPaths = [i for i in self.routes[req.method]]
            for route in regularPaths:
                pattern = re.sub(r"\{([\w]+)\:int\}", r"(?P<\1>[-]?[\\d]+)", route)
                pattern = re.sub(r":([\w]+)", r"(?P<\1>[\\w]+)", pattern)
                pattern = "^" + pattern + "$"
                match = re.match(pattern, req.path)
                if match:
                    handler = self.routes[req.method][route]
                    queryParams = match.groupdict()
                    for i in queryParams:
                        req.params[i] = queryParams[i]
                    def middleware(app, req, res, next):
                        res.status(200)
                        handler(app, req, res)
                        next()
                    stack.append(middleware)
                    break

        stack.reverse()
        next()
        data = str(res).encode(res.encoding or 'latin1') #str(res).encode() if res.encoding is None else str(res).encode(res.encoding)
        clientPort.send(data)

Middleware = Callable[[Application, HttpRequest, HttpResponse, Callable[[], None]], None]
Callback = Callable[[Application, HttpRequest, HttpResponse], None]
ServerCallback = Callable[[HttpRequest, HttpResponse], None]
DictProvider = Callable[[], Dict[str, Any]]
CallbackMutator = Callable[[Callback], Callback]

def build_application(port: int, middlewares: List[Middleware], routers: Dict[str, Router]):
    app = Application(port)
    for m in middlewares:
        app.configure(m)
    for r in routers:
        app.use_router(routers[r], r)
    return app

def create_server(port:int, callback:ServerCallback) -> Application:
    app = Application(port)
    def innerMiddleware(application:Application, req:HttpRequest, res:HttpResponse, next):
        callback(req, res)
        next()
    app.configure(innerMiddleware)
    return app